/*! \file CMT3.h
	\brief	Contains the CMT Level 3 interfaces

	This level contains the high level message interfaces of CMT.
	All code should be OS-independent.

	\date 2006-04-14	Job Mulder, Created
*/

#ifndef CMT3_H_2006_04_14
#define CMT3_H_2006_04_14

#include "CMT2.h"

namespace xsens {

//////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Constants  ///////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

enum OutputMode {
	CALIBRATED_DATA		= 0x0001,
	ORIENTATION_DATA	= 0x0002,
	RAW_DATA			= 0x4000
};

enum OutputSettings {
	SAMPLE_COUNTER		= 0x1,
	QUATERNION			= 0x0,
	EULER				= 0x4,
	MATRIX				= 0x8
};

#define CMT_MAX_DEVICES_PER_PORT		10
#define CMT_DEFAULT_OUTPUT_MODE			OutputMode::ORIENTATION_DATA
#define CMT_DEFAULT_OUTPUT_SETTINGS		(OutputSettings::QUATERNION |\
										 OutputSettings::SAMPLE_COUNTER)
#define CMT_DEFAULT_SAMPLE_FREQUENCY	100.0
#define CMT_DEFAULT_FILTER_GAIN			1.0f
#define CMT_DEFAULT_FILTER_WEIGHTING	1.0f

enum ResetMethod {
	RESET_STORE = 0,
	RESET_HEADING,
	RESET_GLOBAL,
	RESET_OBJECT,
	RESET_ALIGN
};

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Support  classes ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//! The type of a Device Id
typedef unsigned long	DeviceId;

//! A structure for holding deviceId to busId relationships
struct DeviceBusId {
	DeviceId deviceId;
	unsigned char busId;
};

//! A structure for storing device modes
class DeviceMode {
private:
	unsigned short	m_period;
	unsigned short	m_skipFactor;
	unsigned short	m_outputMode;
	unsigned long	m_outputSettings;

public:
	//! default constructor, initializes to the given (default) MT settings
	DeviceMode(	const OutputMode mode = CMT_DEFAULT_OUTPUT_MODE,
				const OutputSettings settings = CMT_DEFAULT_OUTPUT_SETTINGS,
				const double frequency = CMT_DEFAULT_SAMPLE_FREQUENCY);
	
	//! Return the sample frequency in Hz
	double getSampleFrequency(void);
	
	/*! \brief Set the sample frequency in Hz.
	
		This function sets the m_period and m_skipFactor fields.
	*/
	void setSampleFrequency(const double frequency);
	
	//! Return the stored output mode
	OutputMode getOutputMode (void);
	
	//! Set the output mode
	void setOutputMode (const OutputMode mode);
	
	//! Return the stored output settings
	OutputSettings getOutputSettings (void);
	
	//! Set the output settings
	void setOutputSettings (const OutputSettings settings);

};

//! A structure for storing MT filter settings
struct FilterSettings {
	float	m_gain;
	float	m_weighting;
	bool	m_store;
	
	//! Initializes the structure with the given (default) values.
	FilterSettings(	const float gain = CMT_DEFAULT_FILTER_GAIN,
					const float weighting = CMT_DEFAULT_FILTER_WEIGHTING,
					const bool store = false);
};

//! A structure for storing the firmware version
struct FirmwareRevision {
	unsigned char major;
	unsigned char minor;
	unsigned char revision;
};

//! A structure for storing sync in settings
class SyncInSettings {
private:
	unsigned short	m_mode;
	unsigned short	m_skipFactor;
	unsigned long	m_offset;

public:
	//! default constructor, initializes to the given (default) MT settings
	SyncInSettings(	const SyncInMode mode = 0,
					const unsigned short skip = 0,
					const unsigned long offset = 0);
	
	//! Return the stored sync in mode
	SyncInMode getMode (void);
	
	//! Set the sync mode
	void setMode (const SyncInMode mode);
	
	//! Return the stored skip factor
	unsigned short getSkipFactor (void);
	
	//! Set the skip factor
	void setSkipFactor (const unsigned short skip);
	
	//! Return the stored offset in ns
	unsigned long getOffset (void);
	
	//! Set the offset in ns
	void setOffset (const unsigned long offset);
};

//! A structure for storing sync out settings
class SyncOutSettings {
private:
	unsigned short	m_mode;
	unsigned short	m_skipFactor;
	unsigned long	m_offset;
	unsigned long	m_pulseWidth;

public:
	//! default constructor, initializes to the given (default) MT settings
	SyncOutSettings(const SyncOutMode mode = 0,
					const unsigned short skip = 0,
					const unsigned long offset = 0,
					const unsigned long width = 1000000);
	
	//! Return the stored sync out mode
	SyncOutMode getMode (void);
	
	//! Set the sync mode
	void setMode (const SyncOutMode mode);
	
	//! Return the stored skip factor
	unsigned short getSkipFactor (void);
	
	//! Set the skip factor
	void setSkipFactor (const unsigned short skip);
	
	//! Return the stored offset in ns
	unsigned long getOffset (void);
	
	//! Set the offset in ns
	void setOffset (const unsigned long offset);

	//! Return the stored pulse width in ns
	unsigned long getPulseWidth (void);
	
	//! Set the pulse width in ns
	void setPulseWidth (const unsigned long offset);
};

/*! \brief A structure containing MT data.

*/
class CMTData 
{
private:
	unsigned long m_dataLength;
	unsigned char* m_data;
	
public:
	CMTData(const unsigned long length, const unsigned char* data = NULL);
	~CMTData();
	
	//! 
	unsigned char* data(void);
};

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT3s  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief High-level serial communication class.

	The class uses CMT level 2, but does not inherit from it. If software needs to access 
	the level 2 component, it needs to be done through the cmt2s() function.
	
	All device identification is done through device Ids in CMT level 3.
*/
class CMT3s {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT3s(const CMT3s& ref);

protected:
	//! The CMT level 2 object that this class operates on
	CMT2s m_cmt2s;

	CMTResult m_lastResult;			//!< The last result of an operation
	
	/*! A list holding the connected device and bus Ids. The list ends at the first device
		that has Id 0.
	*/
	DeviceBusId m_deviceBusList[CMT_MAX_DEVICES_PER_PORT];
	
public:
	//! Default constructor, initializes all members to their default values.
	CMT3s();

	//! Destructor, de-initializes, frees memory allocated for buffers, etc.
	~CMT3s();

	/*! \brief Close the serial communication port.
	
		Buffered data remains available until the object is destroyed or until the open()
		function is called again.
	*/
	CMTResult close (void);

	/*! \brief Return a reference to the embedded CMT2s object.
	
		Any manipulation of the object should be done through CMT3s. CMT3s integrity is 
		not guaranteed if the CMT2s object is manipulated directly.
	*/
	CMT2s* cmt2s(void);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);

	/*! \brief Wait for a data message to arrive.
	
		The function waits for a data message to arrive or until a timeout occurs. If no
		data message has arrived in time or an error occurs, a NULL pointer is returned.
		The getLastError() function can be used to check what went wrong.
		If the function succeeds, a Message object is returned that contains the data
		message.
	*/
	Message* waitForDataMessage(void);

	//! \brief Return whether the communication port is open or not.
	bool isOpen (void) const;

	/*! \brief Open a communication channel to the given COM port number.
	
		This function is first passed through to the CMT2s object. Then, the device 
		settings are retrieved and stored locally. This function automatically places the 
		device(s) in config mode, using \c gotoConfig().
	*/
	CMTResult open (const unsigned char portNumber,
					const unsigned long baudRate = CMT1S_DEFAULT_BAUD_RATE);

	/*! \brief Set the sizes of the inbound and outbound data buffers.
	
		This function is passed through to the CMT2s object.
	*/
	CMTResult setReadBufferSize (const unsigned long readBufSize);

	/*! \brief Set the default timeout value to use in blocking operations.

		This function is passed through to the CMT2s object.
	*/
	CMTResult setTimeout (const unsigned long ms);

	//! Place all connected sensors into Configuration Mode.
	CMTResult gotoConfig (void);
	//! Place all connected sensors into Measurement Mode.
	CMTResult gotoMeasurement (void);
	
	/*! \brief Retrieve the BusId of a device.
	
		The function checks its internal list for a match of deviceId. If it is found, the
		corresponding BusId is returned in busId. Otherwise, a 0 is placed in busId and a
		NOT_FOUND result is returned.
	*/
	CMTResult getBusId (const DeviceId deviceId, unsigned char& busId);
	
	//! Get the state (enabled/disabled) of the AMD algorithm
	CMTResult getAmdState (const DeviceId deviceId, bool& state);
	//! Set the state (enabled/disabled) of the AMD algorithm
	CMTResult setAmdState (const DeviceId deviceId, const bool state);

	//! Get the baudrate that is currently being used by the port
	CMTResult getBaudrate (const unsigned long& baudrate);
	//! Set the baudrate and reconnect at the new baudrate if succesful.
	CMTResult setBaudrate (const unsigned long baudrate);

	//! Retrieve the complete device configuration of a device.
	CMTResult getConfiguration (const DeviceId deviceId,
								DeviceConfiguration& configuration);

	//! Retrieve the number of bytes that are in a data message as sent by the device.
	CMTResult getDataLength (const DeviceId deviceId, unsigned long& length);

	//! Return the number of connected devices. Returns 0 if not connected.
	unsigned int getDeviceCount (void);

	//! Retrieve the DeviceId of the device at the given BusId 
	CMTResult getDeviceId (const unsigned char busId, DeviceId& deviceId);

	//! Retrieve the complete device output mode of a device.
	CMTResult getDeviceMode (const DeviceId deviceId, DeviceMode& mode);
	//! Set the complete device output mode of a device.
	CMTResult setDeviceMode (const DeviceId deviceId, const DeviceMode& mode);

	//! Retrieve the filter settings of a device.
	CMTResult getFilterSettings (const DeviceId deviceId, FilterSettings& settings);
	//! Set the filter settings of a device.
	CMTResult setFilterSettings (	const DeviceId deviceId,
									const FilterSettings& settings);

	//! Retrieve the firmware revision of a device.
	CMTResult getFirmwareRevision (	const DeviceId deviceId,
									FirmwareRevision& revision);
	
	//! Retrieve the heading offset of a device. The range is -pi to +pi.
	CMTResult getHeading (const DeviceId deviceId, float& heading);
	//! Set the heading offset of a device. The valid range is -pi to +pi.
	CMTResult setHeading (const DeviceId deviceId, const float heading);
	
	//! Retrieve the location ID of a device. The buffer should be at least 20 bytes.
	CMTResult getLocationId (const DeviceId deviceId, char* locationId);
	//! Set the location ID of a device. The buffer should be no more than 20 bytes.
	CMTResult setLocationId (const DeviceId deviceId, const char* locationId);

	//! Retrieve the product code of a device. The buffer should be at least 20 bytes.
	CMTResult getProductCode (const DeviceId deviceId, char* productCode);

	//! Retrieve the inbound synchronization settings of a device.
	CMTResult getSyncInSettings (const DeviceId deviceId, SyncInSettings& settings);
	//! Set the inbound synchronization settings of a device.
	CMTResult setSyncInSettings (	const DeviceId deviceId,
									const SyncInSettings& settings);
	
	//! Retrieve the outbound synchronization settings of a device.
	CMTResult getSyncOutSettings (	const DeviceId deviceId,
									SyncOutSettings& settings);
	//! Set the outbound synchronization settings of a device.
	CMTResult setSyncOutSettings (	const DeviceId deviceId,
									const SyncOutSettings& settings);

	//! Retrieve a data message.
	CMTResult readDataMessage (CMTData& data);
	//! Request a data message and wait for it to arrive.
	CMTResult requestData (const DeviceId deviceId, CMTData& data);
	
	//! Reset all connected sensors.
	CMTResult reset (void);
	
	//! Perform an orientation reset on a device.
	CMTResult resetOrientation(const DeviceId deviceId, const ResetMethod method);
	
	//! Restore the factory defaults of a device.
	CMTResult restoreFactoryDefaults(const DeviceId deviceId);
};

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////// CMT3f  /////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief High-level file communication class.

	The class uses CMT level 2, but does not inherit from it. If software needs to access 
	the level 2 component, it needs to be done through the cmt2f() function.
	
	All device identification is done through device Ids in CMT level 3.
*/
class CMT3f {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT3f(const CMT3f& ref);

protected:
	//! The CMT level 2 object that this class operates on
	CMT2f m_cmt2f;

	CMTResult m_lastResult;			//!< The last result of an operation
	
	/*! A list holding the device Ids that are logged in the file. The list ends at the
		first device that has Id 0.
	*/
	DeviceId m_deviceList[CMT_MAX_DEVICES_PER_PORT];
	
public:
	//! Default constructor, initializes all members to their default values.
	CMT3f();

	//! Destructor, de-initializes, frees memory allocated for buffers, etc.
	~CMT3f();

	//! \brief Close the file.
	CMTResult close (void);

	/*! \brief Return a reference to the embedded CMT1f object.
	
		Any manipulation of the object should be done through CMT3f. CMT3f integrity is 
		not guaranteed if the CMT2f object is manipulated directly.
	*/
	CMT2f* cmt2f (void);

	//! Return the error code of the last operation.
	CMTResult getLastError(void);

	//! \brief Return whether a file is open or not.
	bool isOpen (void) const;

	/*! \brief Open a log-file for reading.
	
		This function is first passed through to the CMT2f object. Then, the device 
		settings are retrieved and stored locally.
	*/
	CMTResult open (const char *filename);

	/*! \brief Retrieve the index of a device.
	
		The function checks its internal list for a match of deviceId. If it is found, the
		index in the list is returned in index. Otherwise, 255 is placed in index and a
		NOT_FOUND result is returned.
	*/
	CMTResult getDeviceIndex (const DeviceId deviceId, unsigned char& index);
	
	//! Get the state (enabled/disabled) of the AMD algorithm. May return NOT_AVAILABLE.
	CMTResult getAmdState (const DeviceId deviceId, bool& state);

	//! Retrieve the complete device configuration of a device.
	CMTResult getConfiguration (const DeviceId deviceId,
								DeviceConfiguration& configuration);

	//! Retrieve the number of bytes that are in a data message as sent by the device.
	CMTResult getDataLength (const DeviceId deviceId, unsigned long& length);

	//! Return the number of connected devices. Returns 0 if file not open.
	unsigned int getDeviceCount (void);

	//! Retrieve the DeviceId of the device at the given index
	CMTResult getDeviceId (const unsigned char index, DeviceId& deviceId);

	//! Retrieve the complete device output mode of a device.
	CMTResult getDeviceMode (const DeviceId deviceId, DeviceMode& mode);

	//! Retrieve the filter settings of a device. May return NOT_AVAILABLE.
	CMTResult getFilterSettings (const DeviceId deviceId, FilterSettings& settings);

	//! Retrieve the firmware revision of a device. May return NOT_AVAILABLE.
	CMTResult getFirmwareRevision (	const DeviceId deviceId,
									FirmwareRevision& revision);
	
	/*! Retrieve the heading offset of a device. The range is -pi to +pi. May return
		NOT_AVAILABLE.
	*/
	CMTResult getHeading (const DeviceId deviceId, float& heading);
	
	/*! Retrieve the location ID of a device. The buffer should be at least 20 bytes. May
		return NOT_AVAILABLE.
	*/
	CMTResult getLocationId (const DeviceId deviceId, char* locationId);

	/*! Retrieve the product code of a device. The buffer should be at least 20 bytes. May
		return NOT_AVAILABLE.
	*/
	CMTResult getProductCode (const DeviceId deviceId, char* productCode);

	/*! Retrieve the inbound synchronization settings of a device. May return
		NOT_AVAILABLE.
	*/
	CMTResult getSyncInSettings (const DeviceId deviceId, SyncInSettings& settings);
	
	/*! Retrieve the outbound synchronization settings of a device. May return
		NOT_AVAILABLE.
	*/
	CMTResult getSyncOutSettings (	const DeviceId deviceId,
									SyncOutSettings& settings);

	//! Retrieve a data message. May return NOT_AVAILABLE if no more data is available.
	CMTResult readDataMessage (CMTData& data);
	
	//! Reset the data rad pointers to the start of the buffers.
	CMTResult reset (void);
};

} // end of xsens namespace

#endif
