/*! \file CMT4.h
	\brief	Contains the CMT Level 4 interfaces

	This level contains the thread level interfaces of CMT.
	All code should be OS-independent.

	\date 2006-04-18	Job Mulder, Created
*/

#ifndef CMT4_H_2006_04_18
#define CMT4_H_2006_04_18

#include "CMT3.h"
//#include <queue>	// an alternative for the Queue class
//typedef std::queue<CMTData*>	CMTDataBuffer;
//typedef std::queue<Message*>	MessageBuffer;

namespace xsens {

//////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////// Constants  ///////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

#define CMT_MAX_PORTS_PER_THREAD	20
#define CMT_MAX_FILES_PER_THREAD	20
#define CMT4_MAX_DEVICES		((CMT_MAX_FILES_PER_THREAD + CMT_MAX_PORTS_PER_THREAD) \
								* CMT_MAX_DEVICES_PER_PORT)

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////// Support  classes ////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//! A class for holding device identification fields.
struct DeviceIdentificator {
	DeviceId	m_deviceId;
	CMT3s*		m_cmt3s;
	CMT3f*		m_cmt3f;
	char 		m_locationId[20];
	
	//! Constructor, sets the deviceID to 0 and both pointers to NULL.
	DeviceIdentificator() : deviceId(0), cmt3s(NULL), cmt3f(NULL), locationId("") {}
		
	//! Copy the given location Id into m_locationId.
	void setLocationId(const char * locationId);
	
	//! Return true if the given locationId is the same as m_locationId, false otherwise.
	bool checkLocationId(const char * locationId);
};

/*! \brief A FIFO queue with limited length (cyclic).

	The class is based on the STL queue class, but has a limited size. If more items are
	inserted than would fit, the oldest item is overwritten.
*/
template <class T>
class Queue<T> {
private:
	unsigned long m_maxCount;
	unsigned long m_currentCount;
	unsigned long m_first;
	
	T*	m_list;
public:
	typedef T	value_type;
	
	typedef T		value_type;	//!< The type of the value stored in this queue.
	typedef size_t	size_type;	//!< The type of a 'size' value.
	Queue<T>(size_type size=16); 	//!< Create an empty queue with capacity size.
	Queue<T>(const Queue<T>& q);		//!< The copy constructor.
	Queue<T>& operator=(const Queue<T>& q);	//!< The assignment operator.
	bool empty() const;					//!< Return true if the queue is empty.
	size_type size() const;				//!< Return the number of elements in the queue.
	value_type& front();				//!< Return the oldest element in the queue.
	const value_type& front() const;	//!< Return the oldest element in the queue.
	value_type& back();					//!< Return the newest element in the queue.
	const value_type& back() const;		//!< Return the newest element in the queue.
	void push(const value_type& x);		//!< Insert x at the back of the queue.
	void pop(void);						//!< Remove the element at the front of the queue.
};

//! A type for storing data packets
typedef Queue<CMTData*>	DataQueue;

//////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// CMT4 //////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

/*! \brief Thread-level CMT class.

	The class uses CMT level 3, but does not inherit from it. If software needs to access 
	the level 3 components, it needs to be done through the cmt3s and cmt3f functions.
	
	All device identification is done through device Ids in CMT level 4.
*/
class CMT4 {
private:
	//! This object cannot be copied, so this function is not implemented.
	CMT4(const CMT4& ref);

protected:
	/*! The CMT level 3 serial objects that this class operates on.
		The list ends with a NULL pointer.
	*/
	CMT3s* m_cmt3s[CMT_MAX_PORTS_PER_THREAD];
	
	/*! A list of queues holding data packets from several serial devices. The list is 
		exactly as long as the m_cmt3s list.
	*/
	DataQueue m_dataQueueList[CMT_MAX_PORTS_PER_THREAD];
	
	/*! The CMT level 3 file objects that this class operates on.
		The list ends with a NULL pointer.
	*/
	CMT3f* m_cmt3f[CMT_MAX_FILES_PER_THREAD];

	/*! The CMT level 2 file objects that this class can log to.
		The list ends with a NULL pointer.
	*/
	CMT2f* m_logFiles[CMT_MAX_FILES_PER_THREAD];

	CMTResult m_lastResult;			//!< The last result of an operation
	
	/*! A list holding the connected device and bus Ids. The list ends at the first device
		that has Id 0.
	*/
	DeviceIdentificator m_deviceIdentificatorList[CMT4_MAX_DEVICES];
	
	// thread management
//	?? m_thread;
//	?? m_lock;
	volatile bool m_isRunning;
	volatile bool m_stopping;

	//! Indicates that the data read thread should temporarily stop reading data.
	volatile bool m_suspending;
	//! Indicates that the data read thread has temporarily stopped reading data.
	volatile bool m_isSuspended;

	/*! Set the m_suspending flag, trigger proper processing of the flag and wait for
		m_suspended.
	*/
	void suspendThread(void);
	
	//! Clear the m_suspended flag and trigger proper processing of the flag.
	void continueThread(void);

public:
	/*! \brief Default constructor, initializes all members to their default values.
	
		The function does not start the thread yet.
	*/
	CMT4();

	/*! Destructor, de-initializes, frees memory allocated for buffers, etc.
	
		If the thread is running, it is stopped and terminated first.
	*/
	~CMT4();

	/*! \brief Close a serial communication port.
	*/
	CMTResult closeSerial (const unsigned long portNr);

	/*! \brief Close an input file.
	*/
	CMTResult closeFile (const char * filename);

	/*! \brief Close all devices.
	*/
	CMTResult close (void);

	/*! \brief Return a reference to an embedded CMT3s object.
	
		Any manipulation of the object should be done through CMT4. CMT4 integrity is 
		not guaranteed if the CMT3s objects are manipulated directly.
	*/
	CMT3s* cmt3s(const unsigned long portNr);

	/*! \brief Return a reference to an embedded CMT3f object.
	
		Any manipulation of the object should be done through CMT4. CMT4 integrity is 
		not guaranteed if the CMT3f objects are manipulated directly.
	*/
	CMT3f* cmt3f(const char * filename);

	//! Return the error code of the last operation.
	CMTResult getLastError(void) const;

	/*! \brief Open a communication channel to the given COM port number.
	
		A new CMT3s object is created and the function is passed through to it. If the
		operation succeeds, the connected devices are placed in configuration mode and the
		level 3 object is stored in the appropriate list. Otherwise, the object is
		immediately destroyed and an error is returned.
	*/
	CMTResult openPort (const unsigned char portNumber,
					const unsigned long baudRate = CMT1S_DEFAULT_BAUD_RATE);

	/*! \brief Open a file.
	
		A new CMT3f object is created and the function is passed through to it. If the
		operation succeeds, the device configuration is read from the file and the level 3
		object is stored in the appropriate list. Otherwise, the object is immediately
		destroyed and an error is returned.
	*/
	CMTResult openFile (const char * filename);

	//! Place all sensors connected through a serial port into Configuration Mode.
	CMTResult gotoConfig (void);
	
	//! Place all sensors connected through a serial port into Measurement Mode.
	CMTResult gotoMeasurement (void);

	//! Return the number of connected devices. Returns 0 if not connected.
	unsigned long getDeviceCount (void) const;

	//! Reset all connected sensors.
	CMTResult reset (void);

	//! Retrieve a data message.
//	CMTResult readDataMessage (CMTData& data);
	
	//! Get the state (enabled/disabled) of the AMD algorithm
	CMTResult getAmdState (const DeviceId deviceId, bool& state);
	//! Set the state (enabled/disabled) of the AMD algorithm
	CMTResult setAmdState (const DeviceId deviceId, const bool state);

	//! Retrieve the complete device configuration of a device.
	CMTResult getConfiguration (const DeviceId deviceId,
								DeviceConfiguration& configuration);

	//! Retrieve the number of bytes that are in a data message as sent by the device.
	CMTResult getDataLength (const DeviceId deviceId, unsigned long& length);

	//! Retrieve the complete device output mode of a device.
	CMTResult getDeviceMode (const DeviceId deviceId, DeviceMode& mode);
	//! Set the complete device output mode of a device.
	CMTResult setDeviceMode (const DeviceId deviceId, const DeviceMode& mode);

	//! Retrieve the filter settings of a device.
	CMTResult getFilterSettings (const DeviceId deviceId, FilterSettings& settings);
	//! Set the filter settings of a device.
	CMTResult setFilterSettings (	const DeviceId deviceId,
									const FilterSettings& settings);

	//! Retrieve the firmware revision of a device.
	CMTResult getFirmwareRevision (	const DeviceId deviceId,
									FirmwareRevision& revision);
	
	//! Retrieve the heading offset of a device. The range is -pi to +pi.
	CMTResult getHeading (const DeviceId deviceId, float& heading);
	//! Set the heading offset of a device. The valid range is -pi to +pi.
	CMTResult setHeading (const DeviceId deviceId, const float heading);
	
	//! Retrieve the location ID of a device. The buffer should be at least 20 bytes.
	CMTResult getLocationId (const DeviceId deviceId, char* locationId);
	//! Set the location ID of a device. The buffer should be no more than 20 bytes.
	CMTResult setLocationId (const DeviceId deviceId, const char* locationId);

	//! Retrieve the product code of a device. The buffer should be at least 20 bytes.
	CMTResult getProductCode (const DeviceId deviceId, char* productCode);

	//! Retrieve the inbound synchronization settings of a device.
	CMTResult getSyncInSettings (const DeviceId deviceId, SyncInSettings& settings);
	//! Set the inbound synchronization settings of a device.
	CMTResult setSyncInSettings (	const DeviceId deviceId,
									const SyncInSettings& settings);
	
	//! Retrieve the outbound synchronization settings of a device.
	CMTResult getSyncOutSettings (	const DeviceId deviceId,
									SyncOutSettings& settings);
	//! Set the outbound synchronization settings of a device.
	CMTResult setSyncOutSettings (	const DeviceId deviceId,
									const SyncOutSettings& settings);

	//! Request a data message and wait for it to arrive.
	CMTResult requestData (const DeviceId deviceId, CMTData& data);

	//! Perform an orientation reset on a device.
	CMTResult resetOrientation(const DeviceId deviceId, const ResetMethod method);
	
	//! Restore the factory defaults of a device.
	CMTResult restoreFactoryDefaults(const DeviceId deviceId);

	
	// Identification by locationId, all functions simply call the deviceId function.

	//! Retrieve the deviceId that matches the given locationId.
	DeviceId getDeviceId(const char * locationId) const;

	//! Get the state (enabled/disabled) of the AMD algorithm
	CMTResult getAmdState (const char * locationId, bool& state)
		{ return getAmdState (getDeviceId(locationId),state); }
		
	//! Set the state (enabled/disabled) of the AMD algorithm
	CMTResult setAmdState (const char * locationId, const bool state)
		{ return setAmdState (getDeviceId(locationId),state); }

	//! Retrieve the complete device configuration of a device.
	CMTResult getConfiguration (const char * locationId,
								DeviceConfiguration& configuration);
		{ return getConfiguration (getDeviceId(locationId),configuration); }

#if 0
#endif

	//! Retrieve the number of bytes that are in a data message as sent by the device.
	CMTResult getDataLength (const char * locationId, unsigned long& length);
		{ return getDataLength (getDeviceId(locationId), length); }

	//! Retrieve the complete device output mode of a device.
	CMTResult getDeviceMode (const char * locationId, DeviceMode& mode);
		{ return getDeviceMode (getDeviceId(locationId), mode);
	//! Set the complete device output mode of a device.
	CMTResult setDeviceMode (const char * locationId, const DeviceMode& mode);
		{ return setDeviceMode (getDeviceId(locationId), mode); }

	//! Retrieve the filter settings of a device.
	CMTResult getFilterSettings (const char * locationId, FilterSettings& settings);
		{ return getFilterSettings (getDeviceId(locationId), settings);
	//! Set the filter settings of a device.
	CMTResult setFilterSettings (	const char * locationId,
									const FilterSettings& settings);
		{ return setFilterSettings (getDeviceId(locationId), settings); }

	//! Retrieve the firmware revision of a device.
	CMTResult getFirmwareRevision (	const char * locationId,
									FirmwareRevision& revision);
		{ return getFirmwareRevision (getDeviceId(locationId), revision); }
	
	//! Retrieve the heading offset of a device. The range is -pi to +pi.
	CMTResult getHeading (const char * locationId, float& heading);
		{ return getHeading (getDeviceId(locationId), heading); }
	//! Set the heading offset of a device. The valid range is -pi to +pi.
	CMTResult setHeading (const char * locationId, const float heading);
		{ return setHeading (getDeviceId(locationId), heading); }
	
	/*! Set the new location ID of a device. The buffer should be no more than 20 bytes.
		This also updates the level 4 record of the device.
	*/
	CMTResult setLocationId (const char * locationId, const char* newLocationId);
		{ return setLocationId (getDeviceId(locationId), newLocationId); }

	//! Retrieve the product code of a device. The buffer should be at least 20 bytes.
	CMTResult getProductCode (const char * locationId, char* productCode);
		{ return getProductCode (getDeviceId(locationId), productCode); }

	//! Retrieve the inbound synchronization settings of a device.
	CMTResult getSyncInSettings (const char * locationId, SyncInSettings& settings);
		{ return getSyncInSettings (getDeviceId(locationId), settings); }
	//! Set the inbound synchronization settings of a device.
	CMTResult setSyncInSettings (	const char * locationId,
									const SyncInSettings& settings);
		{ return setSyncInSettings (getDeviceId(locationId),settings); }
	
	//! Retrieve the outbound synchronization settings of a device.
	CMTResult getSyncOutSettings (	const char * locationId,
									SyncOutSettings& settings);
		{ return getSyncOutSettings (getDeviceId(locationId),settings); }
	//! Set the outbound synchronization settings of a device.
	CMTResult setSyncOutSettings (	const char * locationId,
									const SyncOutSettings& settings);
		{ return setSyncOutSettings (getDeviceId(locationId),settings); }

	//! Request a data message and wait for it to arrive.
	CMTResult requestData (const char * locationId, CMTData& data);
		{ return requestData (getDeviceId(locationId), data); }

	//! Perform an orientation reset on a device.
	CMTResult resetOrientation(const char * locationId, const ResetMethod method);
		{ return resetOrientation(getDeviceId(locationId), method); }
	
	//! Restore the factory defaults of a device.
	CMTResult restoreFactoryDefaults(const char * locationId);
		{ return restoreFactoryDefaults(getDeviceId(locationId)); }

//#endif
};

} // end of xsens namespace

#endif
